﻿#include "MongoDB.h"
#ifdef MONGO_DRIVERS
#include "MongoConsts.h"

#include <QCoreApplication>
#include <QDateTime>
#include <QDebug>
#include <QJsonDocument>
#include <QStack>
#include <QThread>
#include <QTime>
#include <QUuid>
#include <QVariant>

#include <cassert>
#include <variant>
#include <ctime>
#include <cstdlib>
#include <iostream>
#include <string>

#ifndef Q_MOC_RUN
#include <bsoncxx/builder/stream/document.hpp>
#include <bsoncxx/builder/basic/document.hpp>
#include <bsoncxx/builder/basic/kvp.hpp>
#include <bsoncxx/json.hpp>
#include <bsoncxx/stdx/make_unique.hpp>
#include <bsoncxx/builder/basic/array.hpp>
#include <bsoncxx/types.hpp>

#include <mongocxx/client.hpp>
#include <mongocxx/exception/bulk_write_exception.hpp>
#include <mongocxx/exception/error_code.hpp>
#include <mongocxx/exception/logic_error.hpp>
#include <mongocxx/exception/operation_exception.hpp>
#include <mongocxx/exception/server_error_code.hpp>
#include <mongocxx/instance.hpp>
#include <mongocxx/logger.hpp>
#include <mongocxx/options/client.hpp>
#include <mongocxx/uri.hpp>
#endif


using namespace std;
using namespace bsoncxx;
using bsoncxx::builder::basic::kvp;
using bsoncxx::builder::basic::make_document;
using bsoncxx::builder::basic::make_array;
using bsoncxx::builder::stream::document;
using bsoncxx::builder::stream::finalize;
using bsoncxx::type;

namespace {
    template<typename T> bsoncxx::document::view_or_value mk_doc(std::string key, T val)
    {
            return make_document(kvp(key, val));
    }
}

MongoDB::MongoDB():
    QObject(nullptr),
    uri(URI),
    db(DB),
    type_cast_warning(false)
{
}

MongoDB::~MongoDB()
{
}

void MongoDB::mongo_init()
{
    QStringList primary_keys_saved;
    primary_keys_saved << "program_type" << "program_index" << "configuration_name";
    if (!remove_broken_docs(COLLECTION_SAVED, primary_keys_saved)){
        qWarning() << "Remove brocken docs from Mongo Saved failed!";
    }

    QStringList primary_keys_device;
    primary_keys_device << "family" << "serial";
    if (!remove_broken_docs(COLLECTION_DEVICE, primary_keys_device)){
        qWarning() << "Remove brocken docs from Mongo device failed!";
    }

    QStringList primary_keys_history;
    primary_keys_history << "program_type" << "run_index" << "run_number" << "date_of_modification";
    if (!remove_broken_docs(COLLECTION_HISTORY, primary_keys_history)){
        qWarning() << "Remove brocken docs from Mongo Saved failed!";
    }

    if (!remove_duplicates(COLLECTION_SAVED)){
        qWarning() << "Remove duplicates from Mongo Saved failed!";
    }

    if (!remove_duplicates_from_devices(COLLECTION_DEVICE)){
        qWarning() << "Remove duplicates from Mongo Saved failed!";
    }
}

bool MongoDB::try_connection()
{
    QString uri_str = URI;
    std::string uri_std = uri_str.toUtf8().constData();
    try{
        auto uri = mongocxx::uri{uri_std};
        mongocxx::client client{uri};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection saved_config = daq_db[COLLECTION_SAVED.toStdString()];

        bsoncxx::document::value test_doc = make_document(
            kvp("test","test")
        );
        auto result = saved_config.insert_one(std::move(test_doc));

        std::string collection_name = COLLECTION_SAVED.toStdString();
        daq_db[collection_name].delete_many(make_document(kvp("test", "test")));

    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("%1: No Mongo connection").arg(QDateTime::currentDateTime().toString());
        qInfo() << "No mongo !!";
        return false;
    }
    return true;
}

RootConfig MongoDB::read_config(QString program_type, QString program_index, QString config_name)
{
    if(program_type.isEmpty()) {
        qWarning() << "read settings failed. Can't find program type";
        return RootConfig();
    }

    QString quri = URI;
    std::string uri_std = quri.toUtf8().constData();
    try {
        auto uri = mongocxx::uri{uri_std};
        mongocxx::client client{uri};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection collection = daq_db[COLLECTION_SAVED.toStdString()];

        //---read settings from mongo
        mongocxx::cursor cursor = collection.find(
                    make_document(
                        kvp("program_type", program_type.toStdString()),
                        kvp("program_index", program_index.toStdString()),
                        kvp("configuration_name", config_name.toStdString())
                        )
                    );

        bsoncxx::document::view root_doc;
        int count = 0;
        for (const bsoncxx::document::view& doc : cursor) {
            root_doc = doc;
            if(++count > 1) {
                qWarning().noquote().nospace()
                        << QString("MongoDB warning: read config failed. "
                                   "More then one configuration with the same "
                                   "keys in database:");
                qWarning().noquote().nospace()
                        << QString("Program type = %1; program index = %2; "
                                   "configuration name = %3")
                           .arg(program_type).arg(program_index)
                           .arg(config_name);
                break;
            }
        }
        if(count == 1)
            return doc_element_to_root(root_doc);
        if(count==0)
            qWarning() << QString("MongoDB warning: read config failed. Can't find configuration: "
                                  "Program type = %1; program index = %2; configuration name = %3")
                          .arg(program_type).arg(program_index).arg(config_name);
    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo read config warning: %1").arg(str);
    }
    return RootConfig();
}

RootConfig MongoDB::doc_element_to_root(bsoncxx::document::view doc)
{
    RootConfig my_root;

    for (bsoncxx::document::element value_el : doc) {
        auto key = value_el.key().to_string();
        if((value_el.type() != bsoncxx::type::k_oid)&&(value_el.type() != bsoncxx::type::k_document)){
            my_root.set_data(QString::fromStdString(key), bsoncxx_to_qvariant(value_el.get_value()));
        }

        //---my children
        if(value_el.type() == type::k_document){
            RootConfig child_root = doc_element_to_root(value_el.get_document());
            child_root.item_name = QString::fromStdString(value_el.key().to_string());
            my_root.append_child(child_root);
        }
    }
    return my_root;
}

bsoncxx::builder::basic::document MongoDB::root_to_doc_element(const RootConfig &root)
{
    bsoncxx::builder::basic::document doc_to_return;

    const auto &data = root.data_const();
    for(QString field: data.keys()){
        const QVariant &qvar = data[field];

        if(qvar.type() == QVariant::ByteArray) {
            const QByteArray &byte_array = qvar.toByteArray();
            bsoncxx::types::b_binary tmp_b;
            tmp_b.bytes = reinterpret_cast<const uint8_t*>(byte_array.data());
            tmp_b.size = byte_array.size();
            doc_to_return.append(kvp(field.toStdString(), tmp_b));
        } else if(qvar.type() == QVariant::String) {
            QByteArray tt = qvar.toString().toUtf8();
            bsoncxx::types::b_utf8 tmp{tt};
            doc_to_return.append(kvp(field.toStdString(), tmp));
        } else {
            doc_to_return.append(
                        kvp(field.toStdString(), qvariant_to_bsoncxx(qvar))
                        );
        }
    }

    for(QString child_name : root.children().keys()){
        bsoncxx::builder::basic::document child_doc = root_to_doc_element(root.children()[child_name]);
        doc_to_return.append(
            kvp(child_name.toStdString(), child_doc)
        );
    }

    return doc_to_return;
}

bool MongoDB::write_config(const RootConfig &root)
{
    RootConfig empty_root = RootConfig();
    QString program_type = root.get_data<QString>("program_type");
    QString program_index = root.get_data<QString>("program_index");
    QString configuration_name = root.get_data<QString>("configuration_name");

    if(program_type.isEmpty()){
        qWarning() << "write config failed. Can't find program type";
        return false;
    }
    if(configuration_name.isEmpty()){
        configuration_name = PROGRAM_CONFIG_DEFAULT;
    }

    QString quri = URI;
    std::string uri_std = quri.toUtf8().constData();
    try {
        auto uri = mongocxx::uri{uri_std};
        mongocxx::client client{uri};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection collection = daq_db[COLLECTION_SAVED.toStdString()];

        delete_config(program_type, program_index, configuration_name);
        if(!root_to_mongo(COLLECTION_SAVED, root)){
            qWarning() << "Write settings to Mongo failed.";
            return false;
        }

        if(type_cast_warning){
            qWarning() << "Type cast warning!!";
            return false;
        }

    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo write config warning: %1").arg(str);
        return false;
    }
    return true;
}

RootConfig MongoDB::read_history_config(QString program_type, QString run_index, int run_number)
{
    if(program_type.isEmpty()){
        qWarning() << "Read history config failed: can't find program type";
        return RootConfig();
    }

    QString quri = URI;
    std::string uri_std = quri.toUtf8().constData();
    try {
        auto uri = mongocxx::uri{uri_std};
        mongocxx::client client{uri};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection collection = daq_db[COLLECTION_HISTORY.toStdString()];

        //---read settings from mongo
        mongocxx::cursor cursor = collection.find(
                    make_document(
                        kvp("program_type", program_type.toStdString()),
                        kvp("run_index", run_index.toStdString()),
                        kvp("run_number", run_number)
                        )
                    );

        bsoncxx::document::view root_doc;
        bool found = false;
        for (const bsoncxx::document::view& doc : cursor) {
            if(found) {
                qWarning().noquote().nospace() << QString("MongoDB warning: read history config failed. More then one configuration with the same keys in database:");
                qWarning().noquote().nospace() << QString("Program type = %1; run index = %2; run number = %3")
                                                  .arg(program_type).arg(run_index).arg(run_number);
                return RootConfig();
            }
            root_doc = doc;
            found = true;
        }

        if(found) {
            return doc_element_to_root(root_doc);
        } else {
            qWarning() << "MongoDB warning: read config failed. Can't find configuration.";
        }
    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo read history config warning: %1").arg(str);
    }
    return RootConfig();
}

bool MongoDB::root_to_mongo(QString collection_name, RootConfig root)
{
    bsoncxx::builder::basic::document temp_doc = root_to_doc_element(root);

    try {
        mongocxx::client client{mongocxx::uri{URI.toUtf8().constData()}};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection collection = daq_db[collection_name.toStdString()];
        const auto result = collection.insert_one(temp_doc.view());
        const auto &bulk_write_res = result->result();
        if(bulk_write_res.inserted_count() == 1) {
            return true;
        }
        qWarning() << QString("%1 warning: inserted_count(%2) != 1")
                      .arg(__PRETTY_FUNCTION__)
                      .arg(bulk_write_res.inserted_count());
    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo root to mongo warning: %1").arg(str);
    }
    return false;
}

QMap<QString, QStringList> MongoDB::get_config_map(QString program_type)
{
    QMap<QString, QStringList> config_map;
    try {
        mongocxx::client client{mongocxx::uri{URI.toUtf8().constData()}};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection collection = daq_db[COLLECTION_SAVED.toStdString()];

        QString name1 = "program_index";
        QString name2 = "configuration_name";
        QString name3 = "delete_datetime";

        mongocxx::options::find opt{};
        opt.projection(
            make_document(
                kvp(name1.toStdString(), 1),
                kvp(name2.toStdString(), 1),
                kvp(name3.toStdString(), 1)
            )
        );

        mongocxx::cursor cursor = collection.find(
            make_document(
                kvp("program_type", program_type.toStdString())
            ),
            opt
        );

        for (const bsoncxx::document::view& doc : cursor) {
            bsoncxx::document::element delete_datetime_element = doc[name3.toStdString()];
            if(delete_datetime_element.length() != 0){
                qint64 delete_datetime_int = delete_datetime_element.get_date().to_int64();
                QDateTime delete_datetime_mongo = QDateTime::fromMSecsSinceEpoch(delete_datetime_int);
                if(!delete_datetime_mongo.isNull())
                    continue;
            }
            bsoncxx::document::element program_index_element = doc[name1.toStdString()];
            std::string program_index_std = program_index_element.get_utf8().value.to_string();
            QString program_or_run_index = QString::fromStdString(program_index_std);

            bsoncxx::document::element configuration_name_element = doc[name2.toStdString()];
            std::string configuration_name_std = configuration_name_element.get_utf8().value.to_string();
            QString configuration_name = QString::fromStdString(configuration_name_std);
            config_map[program_or_run_index].append(configuration_name);
        }
        return config_map;
    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo get config map warning: %1").arg(str);
        return config_map;
    }
}

QMap<QString, QMap<QString, RootConfig> > MongoDB::get_rootconfig_map(QString program_type)
{
    QMap<QString, QMap<QString, RootConfig>> to_return;
    if(program_type.isEmpty())
        return to_return;

    try {
        QString quri = URI;
        std::string uri_std = quri.toUtf8().constData();
        auto uri = mongocxx::uri{uri_std};
        mongocxx::client client{uri};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection collection = daq_db[COLLECTION_SAVED.toStdString()];

        mongocxx::cursor cursor = collection.find(
            make_document(
                kvp("program_type", program_type.toStdString())
            )
        );

        bsoncxx::document::view root_doc;
        RootConfig root;
        for(const bsoncxx::document::view& doc : cursor) {
            root = doc_element_to_root(doc);
            QString program_index = root.get_data<QString>("program_index");
            QString config_name = root.get_data<QString>("configuration_name");
            to_return[program_index][config_name] = root;
        }
    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo get rootconfig map warning: %1").arg(str);
    }
    return to_return;
}

QStringList MongoDB::get_types_list(QString collection_name)
{
    QStringList to_return;
    try {
        mongocxx::client client{mongocxx::uri{URI.toUtf8().constData()}};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection collection = daq_db[collection_name.toStdString()];
        QMap<QString, QStringList> config_map;

        mongocxx::options::find opt{};
        opt.projection(
            make_document(
                kvp("program_type", 1)
            )
        );

        mongocxx::cursor cursor = collection.find(
            make_document(),
            opt
        );

        for (const bsoncxx::document::view& doc : cursor) {
            bsoncxx::document::element program_type_element = doc["program_type"];
            std::string program_type_std = program_type_element.get_utf8().value.to_string();
            QString program_type = QString::fromStdString(program_type_std);
            to_return.append(program_type);
        }
        to_return.removeDuplicates();
        return to_return;
    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo get types list warning: %1").arg(str);
        return to_return;
    }
}

QDateTime MongoDB::get_date_of_modification(QString program_type, QString program_index, QString configuration_name)
{
    QDateTime date_of_modification_mongo;

    try {
        mongocxx::client client{mongocxx::uri{URI.toUtf8().constData()}};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection collection = daq_db[COLLECTION_SAVED.toStdString()];
        QDateTime to_return;

        mongocxx::options::find opt{};
        opt.projection(
            make_document(
                kvp("date_of_modification", 1)
            )
        );

        mongocxx::cursor cursor = collection.find(
            make_document(
                kvp("program_type", program_type.toStdString()),
                kvp("program_index", program_index.toStdString()),
                kvp("configuration_name", configuration_name.toStdString())
            ),
            opt
        );

        for (const bsoncxx::document::view& doc : cursor) {
            bsoncxx::document::element date_of_modification_element = doc["date_of_modification"];
            if(date_of_modification_element.length() != 0){
                qint64 date_of_modification_int = date_of_modification_element.get_date().to_int64();
                date_of_modification_mongo = QDateTime::fromMSecsSinceEpoch(date_of_modification_int);
            }

        }
        return date_of_modification_mongo;

    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo get date of creation warning: %1").arg(str);
        return date_of_modification_mongo;
    }
}

bool MongoDB::delete_config(QString program_type, QString program_index, QString config_name)
{
    QString quri = URI;
    std::string uri_std = quri.toUtf8().constData();

    try {
        auto uri = mongocxx::uri{uri_std};
        mongocxx::client client{uri};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection collection = daq_db[COLLECTION_SAVED.toStdString()];

        collection.delete_many(
            make_document(
                kvp("program_type", program_type.toStdString()),
                kvp("program_index", program_index.toStdString()),
                kvp("configuration_name", config_name.toStdString())
            )
        );
        return true;
    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo delete config warning: %1").arg(str);
        return false;
    }
}

bool MongoDB::write_history_config(RootConfig root)
{
    try {
        if(!root_to_mongo(COLLECTION_HISTORY, root)){
            qWarning() << "Write settings to Mongo failed.";
            return false;
        }
    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo write history config warning: %1").arg(str);
        return false;
    }
    return true;
}

int MongoDB::check_run_number(QString program_type, QString run_index, int run_number)
{
    QString quri = URI;
    std::string uri_std = quri.toUtf8().constData();
    QVector<int> run_numbers;
    try {
        auto uri = mongocxx::uri{uri_std};
        mongocxx::client client{uri};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection collection = daq_db[COLLECTION_HISTORY.toStdString()];
        mongocxx::pipeline p{};
        p.match(mk_doc("program_type", program_type.toStdString()));
        p.match(mk_doc("run_index", run_index.toStdString()));
        p.match(mk_doc("run_number", mk_doc("$gte", run_number)));
        p.project(make_document(kvp("run_number", 1)));
        p.sort(mk_doc("run_number", 1));
        auto cursor = collection.aggregate(p, mongocxx::options::aggregate{});
        for (const bsoncxx::document::view& doc : cursor) {
            bsoncxx::document::element run_number_element = doc["run_number"];
            int run_number = run_number_element.get_int32();
            run_numbers.append(run_number);
        }
    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo check run number warning: %1").arg(str);
        return -1;
    }

    if(run_numbers.size() ==0)
        return run_number;
    if(run_numbers.size() != 0)
        if(run_number != run_numbers[0]){
            return run_number;
        }

    int k = run_number;
    for(int i=0; i< run_numbers.size(); ++i){
        if(run_numbers[i]-k > 1){
            break;
        }
        k = run_numbers[i];
    }
    return k+1;
}

QList<RootConfig> MongoDB::read_all_config(QString collection_name)
{
    QList<RootConfig> root_config_list;
    QString quri = URI;
    std::string uri_std = quri.toUtf8().constData();
    try {
        auto uri = mongocxx::uri{uri_std};
        mongocxx::client client{uri};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection collection = daq_db[collection_name.toStdString()];
        mongocxx::cursor cursor = collection.find({});
        for (const bsoncxx::document::view& doc : cursor){
            root_config_list.append(doc_element_to_root(doc));
        }
        return root_config_list;
    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo read all config warning: %1").arg(str);
        return root_config_list;
    }
}

bool MongoDB::write_device_config(RootConfig source_root)
{
    quint64 serial_quint64 = source_root.get_data<quint64>("serial", 0);
    if(serial_quint64 == 0) {
        qWarning() << "serial_quint64 is empty";
        return false;
    }
    long serial_int64 = static_cast<long>(serial_quint64);
    QString serial_str = quint64_to_serial_str(serial_quint64);

    QString family = source_root.get_data<QString>("family", "");
    QString target_id = family + "_" + serial_str;

    RootConfig current_from_source = source_root.child("current");
    if(!current_from_source.isEmpty()){
        QDateTime current_dt = QDateTime::currentDateTime();
        current_from_source.set_data("date_from", current_dt);
        source_root.append_child(current_from_source);
    }

    QString quri = URI;
    std::string uri_std = quri.toUtf8().constData();
    try {
        auto uri = mongocxx::uri{uri_std};
        mongocxx::client client{uri};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection collection = daq_db[COLLECTION_DEVICE.toStdString()];

        mongocxx::cursor cursor = collection.find(
            make_document(
                kvp("family", family.toStdString()),
                kvp("serial", serial_int64)
            )
        );

        int count = 0;
        bsoncxx::document::view root_doc;
        for (const bsoncxx::document::view& doc : cursor) {
            root_doc = doc;
            if(++count > 1){
                qWarning().noquote().nospace() << QString("MongoDB warning: write to device config failed. More then one configuration with the same keys in database:");
                qWarning().noquote().nospace() << QString("family = %1; serial = %2")
                                                  .arg(family).arg(serial_str);
                return false;
            }
        }

        if(count == 0){
            return root_to_mongo(COLLECTION_DEVICE, source_root);
        }

        // samiy slozhniy sluchay pognali
        RootConfig target_root =  doc_element_to_root(root_doc);

        RootConfig current_root = target_root.child("current");
        if(!current_root.isEmpty()){
            target_root = replace_current_to_log(target_root);
        }

        source_root.append_child(target_root.child("log"));
        collection.delete_many(
            make_document(
                kvp("family", family.toStdString()),
                kvp("serial", serial_int64)
            )
        );

        return root_to_mongo(COLLECTION_DEVICE, source_root);
    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo write device config warning: %1").arg(str);
        return false;
    }
}

RootConfig MongoDB::replace_current_to_log(RootConfig target_root)
{
    RootConfig current_root = target_root.child("current");
    RootConfig log_root = target_root.child("log");
    //find max and insert
    int log_number_max=0;
    for(QString child_name : target_root.child("log").children().keys()){
        if(child_name.startsWith("log_")){
            QString log_number_str = child_name.split("_").at(1);
            bool ok= false;
            int log_number = log_number_str.toInt(&ok);
            if(ok && (log_number >= log_number_max)){
                log_number_max = log_number;
            }
        }
    }

    RootConfig log_xx = current_root;
    log_xx.id = QUuid::createUuid();
    QDateTime current_dt = QDateTime::currentDateTime();
    log_xx.set_data("date_to", current_dt);
    log_xx.item_name = "log_" + QString("%1").arg(log_number_max+1, 4, 10, QChar('0'));
    log_root.append_child(log_xx);
    target_root.append_child(log_root);
    return target_root;
}

QString MongoDB::quint64_to_serial_str(quint64 value)
{
    QString serial_str = QString("%1-%2").arg((value>>16) & 0xFFFFu, 4, 16, QChar('0'))
            .arg(value&0xFFFFu, 4, 16, QChar('0'));
    if(value>>32) {
        serial_str.prepend(QString("%1-").arg((value>>32) & 0xFFFFu, 4, 16, QChar('0')));
    }
    if(value>>48) {
        serial_str.prepend(QString("%1-").arg((value>>48) & 0xFFFFu, 4, 16, QChar('0')));
    }
    serial_str = serial_str.toUpper();

    return serial_str;
}

quint64 MongoDB::serial_str_to_quint64(QString serial_str)
{
    QString tmp = serial_str.split("-").at(0) + serial_str.split("-").at(1);
    bool ok=  false;
    quint64 value = tmp.toULongLong(&ok, 16);

    if(!ok){
        qWarning() << "MongoDB: Device index serial string to hex failed";
    }
    return value;
}

//---remove docs from mongo
bool MongoDB::remove_broken_docs(QString collection_name, QStringList primary_keys)
{
    try {
        auto uri = mongocxx::uri{URI.toUtf8().constData()};
        mongocxx::client client{uri};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection collection = daq_db[collection_name.toStdString()];

        bsoncxx::builder::basic::array pk_array;
        for(int i=0; i<primary_keys.size(); ++i){
            pk_array.append(
                make_document(
                    kvp(
                        primary_keys[i].toStdString(),
                        make_document(
                            kvp("$exists", false)
                        )
                    )
                )
            );
        }

        try {
            collection.delete_many(
                make_document(
                    kvp(
                        "$or",
                          pk_array
                    )
                )
            );
        } catch (const mongocxx::exception& e){

        }
        return true;
    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo remove broken docs warning: %1").arg(str);
        return false;
    }
}

bool MongoDB::remove_duplicates(QString collection_name)
{
    try {
        auto uri = mongocxx::uri{URI.toUtf8().constData()};
        mongocxx::client client{uri};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection last_config = daq_db[collection_name.toStdString()];

        mongocxx::pipeline p{};

        p.group(make_document(
            kvp("_id",
                make_document(
                    kvp("configuration_name", "$configuration_name"),
                    kvp("program_type", "$program_type"),
                    kvp("program_index", "$program_index")
                )
            ),
            kvp("id", make_document(kvp("$first", "$_id"))),
            kvp("program_type", make_document(kvp("$first", "$program_type"))),
            kvp("program_index", make_document(kvp("$first", "$program_index"))),
            kvp("configuration_name", make_document(kvp("$first", "$configuration_name"))),
            kvp("date_of_modification", make_document(kvp("$max", "$date_of_modification")))
        ));

        QVector<QString> not_dublicated_docs;
        mongocxx::cursor cursor_aggregate = last_config.aggregate(p, mongocxx::options::aggregate{});
        for (const bsoncxx::document::view& doc : cursor_aggregate){
            QString id;
            QString program_type;
            QString program_index;
            QString date_of_modification;

            bsoncxx::document::element id_element = doc["id"];
            std::string oid= id_element.get_oid().value.to_string();
            id = QString::fromStdString(oid);
            not_dublicated_docs.append(id);
        }

        mongocxx::options::find opt{};
        opt.projection(
            make_document(
                kvp("_id", 1)
            )
        );

        try {
            mongocxx::cursor cursor_find = last_config.find({}, opt);
            for (const bsoncxx::document::view& doc : cursor_find){
                bsoncxx::document::element id_element = doc["_id"];
                std::string oid= id_element.get_oid().value.to_string();
                QString id = QString::fromStdString(oid);
                if(!not_dublicated_docs.contains(id)){
                    last_config.delete_one(make_document(kvp("_id", doc["_id"].get_value())));
                    qInfo() << QString("Find duplicates in Mongo %1 success ").arg(collection_name);
                }
            }
        } catch (const mongocxx::exception& e){
            QString str = QString::fromStdString(e.what());
            qInfo() << QString("Find duplicates fails in Mongo collection: %1 : %2").arg(collection_name).arg(str);
            return false;
        }
        return true;
    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo remove duplicates warning: %1").arg(str);
        return false;
    }
}

bool MongoDB::remove_duplicates_from_devices(QString collection_name)
{
    try {
        auto uri = mongocxx::uri{URI.toUtf8().constData()};
        mongocxx::client client{uri};
        mongocxx::database daq_db = client[DB.toStdString()];
        mongocxx::collection device_config = daq_db[collection_name.toStdString()];

        mongocxx::pipeline p{};

        p.group(make_document(
            kvp("_id",
                make_document(
                    kvp("family", "$family"),
                    kvp("serial", "$serial")
                )
            ),
            kvp("id", make_document(kvp("$first", "$_id"))),
            kvp("family", make_document(kvp("$first", "$family"))),
            kvp("serial", make_document(kvp("$first", "$program_index")))
        ));

        QVector<QString> not_dublicated_docs;
        mongocxx::cursor cursor_aggregate = device_config.aggregate(p, mongocxx::options::aggregate{});
        for (const bsoncxx::document::view& doc : cursor_aggregate){
            QString id;
            QString family;
            QString serial;

            bsoncxx::document::element id_element = doc["id"];
            std::string oid= id_element.get_oid().value.to_string();
            id = QString::fromStdString(oid);
            not_dublicated_docs.append(id);
        }

        mongocxx::options::find opt{};
        opt.projection(
            make_document(
                kvp("_id", 1)
            )
        );

        try {
            mongocxx::cursor cursor_find = device_config.find({}, opt);
            for (const bsoncxx::document::view& doc : cursor_find){
                bsoncxx::document::element id_element = doc["_id"];
                std::string oid= id_element.get_oid().value.to_string();
                QString id = QString::fromStdString(oid);
                if(!not_dublicated_docs.contains(id)){
                    device_config.delete_one(make_document(kvp("_id", doc["_id"].get_value())));
                    qInfo() << QString("Find duplicates in Mongo %1 success ").arg(collection_name);
                }
            }
        } catch (const mongocxx::exception& e){
            QString str = QString::fromStdString(e.what());
            qInfo() << QString("Find duplicates fails in Mongo collection : %1 : %2").arg(collection_name).arg(str);
            return false;
        }
        return true;
    } catch (const mongocxx::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo remove duplicates from devices warning: %1").arg(str);
        return false;
    }
}

QVariant MongoDB::bsoncxx_to_qvariant(types::value value)
{
    QVariant q_var;
    bool type_inside = false;

    if(value.type() == bsoncxx::type::k_utf8){
        stdx::string_view view = value.get_utf8().value;
        std::string name = view.to_string();
        QString name_str = QString::fromStdString(name);
        q_var = name_str;
        type_inside = true;
    }

    if(value.type() == bsoncxx::type::k_int32){
        int value_int = value.get_int32();
        q_var = value_int;
        type_inside = true;
    }

    if(value.type() == bsoncxx::type::k_int64){
        qint64 value_int = value.get_int64();
        q_var = value_int;
        type_inside = true;
    }

    if(value.type() == bsoncxx::type::k_double){
        double value_double = value.get_double();
        q_var = value_double;
        type_inside = true;
    }

    if(value.type() == bsoncxx::type::k_bool){
        bool value_bool = value.get_bool();
        q_var = value_bool;
        type_inside = true;
    }

    if(value.type() == bsoncxx::type::k_date){
        qint64 value_date = value.get_date().to_int64();
        QDateTime value_dt = QDateTime::fromMSecsSinceEpoch(value_date);
        q_var = value_dt;
        type_inside = true;
    }

    if(value.type() == bsoncxx::type::k_binary){
        bsoncxx::types::b_binary binary_data = value.get_binary();
        QByteArray value_binary(reinterpret_cast<const char *>(binary_data.bytes), binary_data.size);
        q_var = value_binary;
        type_inside = true;
    }

    if(!type_inside){
        qWarning("Can't convert from bsoncxx::types to QVariant. Please, add type cast");
    }

    return q_var;
}

types::value MongoDB::qvariant_to_bsoncxx(QVariant qvar)
{
    switch(qvar.type()){
        case QVariant::Bool:{
            bsoncxx::types::b_bool tmp;
            tmp.value = qvar.toBool();
            return bsoncxx::types::value(tmp);
        }
        case QVariant::Int:{
            bsoncxx::types::b_int32 tmp;
            tmp.value = qvar.toInt();
            return bsoncxx::types::value(tmp);
        }
        case QVariant::UInt:{
            bsoncxx::types::b_int32 tmp;
            tmp.value = qvar.toInt();
            return bsoncxx::types::value(tmp);
        }
        case QVariant::LongLong:{
            bsoncxx::types::b_int64 tmp;
            tmp.value = qvar.toLongLong();
            return bsoncxx::types::value(tmp);
        }
        case QVariant::ULongLong:{
            bsoncxx::types::b_int64 tmp;
            quint64 tmp_quint = qvar.toULongLong();
            long int temp_std = tmp_quint;
            if(tmp_quint > std::numeric_limits<qint64>::max()){
                qWarning() <<"type cast from qvariant to bsoncxx failed, quint64 is too large!";
            }
            tmp.value = temp_std;
            return bsoncxx::types::value(tmp);
        }
        case QVariant::Double:{
            bsoncxx::types::b_double tmp;
            tmp.value = qvar.toDouble();
            return bsoncxx::types::value(tmp);
        }
        case QVariant::Date:{
            QDate tmp1 = qvar.toDate();
            QDateTime tmp = QDateTime(tmp1);
            bsoncxx::types::b_date value_date_b_date = bsoncxx::types::b_date(std::chrono::system_clock::from_time_t(tmp.toSecsSinceEpoch()));
            return bsoncxx::types::value(value_date_b_date);
        }
        case QVariant::DateTime:{
            QDateTime tmp = qvar.toDateTime();
            bsoncxx::types::b_date value_date_b_date = bsoncxx::types::b_date(std::chrono::system_clock::from_time_t(tmp.toSecsSinceEpoch()));
            return bsoncxx::types::value(value_date_b_date);
        }

        case 35:{
            bsoncxx::types::b_int64 tmp;
            tmp.value = qvar.toULongLong();
            return bsoncxx::types::value(tmp);
        }
        case 36:{
            bsoncxx::types::b_int32 tmp;
            tmp.value = qvar.toInt();
            return bsoncxx::types::value(tmp);
        }
        case 37:{
            bsoncxx::types::b_int32 tmp;
            tmp.value = qvar.toInt();
            return bsoncxx::types::value(tmp);
        }
        case QVariant::ByteArray:
        case QVariant::String:
        default:{
            qWarning() << "Can't convert from QVariant " << qvar <<  "to bsoncxx::types. Please, add type cast";
            bsoncxx::types::b_bool tmp;
            tmp.value = false;
            return bsoncxx::types::value(tmp);
        }
    }
}

void MongoDB::root_debug_print(RootConfig tree_item)
{
    qInfo() << "===ROOT DEBUG PRINT===";

    qInfo() << tree_item.id.toString();
    for(QString data_field : tree_item.data().keys()){
        if (QMetaType::type(tree_item.data().value(data_field).typeName()) == QMetaType::Int){
            qInfo() << "MY FIELD:"<< data_field << tree_item.get_data<int>(data_field);
        }
        if (QMetaType::type(tree_item.data().value(data_field).typeName()) == QMetaType::QString){
            qInfo() << "MY FIELD:"<< data_field << tree_item.get_data<QString>(data_field);
        }
        if (QMetaType::type(tree_item.data().value(data_field).typeName()) == QMetaType::Bool){
           qInfo() << "MY FIELD:" << data_field << tree_item.get_data<bool>(data_field);
        }

        if (QMetaType::type(tree_item.data().value(data_field).typeName()) == QMetaType::QByteArray){
            qInfo() << "MY FIELD:" << data_field << tree_item.get_data<QByteArray>(data_field);
        }

        if (QMetaType::type(tree_item.data().value(data_field).typeName()) == QMetaType::QDateTime){
            qInfo() << "MY FIELD:" << data_field << tree_item.get_data<QDateTime>(data_field);
        }
    }

    for(QString client : tree_item.children().keys()){
        qInfo() << "MY CHILD:" << client;
        root_debug_print(tree_item.children().value(client));
    }
}

void MongoDB::rename_date_of_creation()
{
    mongocxx::client client{mongocxx::uri{URI.toUtf8().constData()}};
    mongocxx::database daq_db = client[DB.toStdString()];
    mongocxx::collection collection = daq_db[COLLECTION_SAVED.toStdString()];

    collection.update_many(
        make_document(),
        make_document(
            kvp("$rename",
                make_document(
                    kvp("date_of_creation","date_of_modification")
                )
            )
        )
     );
}

#endif

//=== GARBAGE ===

